from calendar import timegm
from datetime import datetime, timedelta

import pytest
import morepath
from morepath import Identity
from more.jwtauth import (
    JWTIdentityPolicy,
    verify_refresh_request,
    InvalidTokenError,
    DecodeError,
    ExpiredSignatureError,
    MissingRequiredClaimError,
)
from more.jwtauth.utils import handler
from webtest import TestApp as Client


def test_handler():
    assert handler(None) is None

    refresh_nonce_handler = handler(
        "more.jwtauth.tests.handler.refresh_nonce_handler"
    )
    assert refresh_nonce_handler(None, "user") == "__user__"

    with pytest.raises(ImportError) as excinfo:
        handler("refresh_nonce_handler")
    assert "Could not import the name: refresh_nonce_handler" in str(
        excinfo.value
    )


def test_create_claims_with_refresh_until_and_nonce():
    refresh_nonce_handler = "more.jwtauth.tests.handler.refresh_nonce_handler"

    identity_policy = JWTIdentityPolicy(
        master_secret="secret",
        allow_refresh=True,
        refresh_delta=timedelta(seconds=2),
        refresh_nonce_handler=refresh_nonce_handler,
    )

    userid = "user"
    claims_set = identity_policy.create_claims_set(None, userid)

    now = timegm(datetime.utcnow().utctimetuple())

    assert claims_set["refresh_until"] >= now + 1
    assert claims_set["refresh_until"] <= now + 3
    assert claims_set["nonce"] == "__user__"


def test_refresh_token():
    class App(morepath.App):
        pass

    class Refresh:
        pass

    @App.identity_policy()
    def get_identity_policy(settings):
        jwtauth_settings = settings.jwtauth.__dict__.copy()
        return JWTIdentityPolicy(**jwtauth_settings)

    @App.path(model=Refresh, path="refresh")
    def get_refresh():
        return Refresh()

    @App.json(model=Refresh)
    def refresh(self, request):
        userid = verify_refresh_request(request)

        @request.after
        def remember(response):
            identity = Identity(userid)
            request.app.remember_identity(response, request, identity)

        return {"userid": userid}

    refresh_nonce_handler = "more.jwtauth.tests.handler.refresh_nonce_handler"
    refresh_delta = 3600

    @App.setting_section(section="jwtauth")
    def get_jwtauth_settings():
        return {
            "master_secret": "secret",
            "allow_refresh": True,
            "refresh_delta": refresh_delta,
            "refresh_nonce_handler": refresh_nonce_handler,
        }

    morepath.commit(App)
    app = App()

    settings = app.settings.jwtauth.__dict__.copy()
    identity_policy = JWTIdentityPolicy(**settings)

    now = timegm(datetime.utcnow().utctimetuple())

    claims_set = {
        "sub": "user",
        "refresh_until": now + refresh_delta,
        "nonce": "__user__",
    }

    token = identity_policy.encode_jwt(claims_set)
    headers = {"Authorization": "JWT " + token}

    c = Client(app)

    r = c.get("/refresh", headers=headers)

    assert r.json == {
        "userid": "user",
    }

    authtype, token = r.headers["Authorization"].split(" ", 1)
    claims_set_decoded = identity_policy.decode_jwt(token)

    assert identity_policy.get_userid(claims_set_decoded) == "user"


def test_refresh_nonce_handler_set_by_decorator():
    class App(morepath.App):
        pass

    class Refresh:
        pass

    @App.identity_policy()
    def get_identity_policy(settings):
        jwtauth_settings = settings.jwtauth.__dict__.copy()
        return JWTIdentityPolicy(**jwtauth_settings)

    @App.path(model=Refresh, path="refresh")
    def get_refresh():
        return Refresh()

    @App.json(model=Refresh)
    def refresh(self, request):
        userid = verify_refresh_request(request)

        @request.after
        def remember(response):
            identity = Identity(userid)
            request.app.remember_identity(response, request, identity)

        return {"userid": userid}

    refresh_delta = 3600

    @App.setting(section="jwtauth", name="refresh_nonce_handler")
    def get_handler():
        def refresh_nonce_handler(request, userid):
            return "__" + userid + "__"

        return refresh_nonce_handler

    @App.setting_section(section="jwtauth")
    def get_jwtauth_settings():
        return {
            "master_secret": "secret",
            "allow_refresh": True,
            "refresh_delta": refresh_delta,
        }

    morepath.commit(App)
    app = App()
    c = Client(app)

    settings = app.settings.jwtauth.__dict__.copy()
    identity_policy = JWTIdentityPolicy(**settings)

    now = timegm(datetime.utcnow().utctimetuple())

    claims_set = {
        "sub": "user",
        "refresh_until": now + refresh_delta,
        "nonce": "__user__",
    }

    token = identity_policy.encode_jwt(claims_set)
    headers = {"Authorization": "JWT " + token}

    r = c.get("/refresh", headers=headers)

    assert r.json == {
        "userid": "user",
    }

    authtype, token = r.headers["Authorization"].split(" ", 1)
    claims_set_decoded = identity_policy.decode_jwt(token)

    assert identity_policy.get_userid(claims_set_decoded) == "user"


def test_refresh_token_with_extra_claims():
    class App(morepath.App):
        pass

    class Refresh:
        pass

    @App.identity_policy()
    def get_identity_policy(settings):
        jwtauth_settings = settings.jwtauth.__dict__.copy()
        return JWTIdentityPolicy(**jwtauth_settings)

    @App.path(model=Refresh, path="refresh")
    def get_refresh():
        return Refresh()

    @App.json(model=Refresh)
    def refresh(self, request):
        userid = verify_refresh_request(request)
        updated_extra_claims = {
            "fullname": "Harry Potter",
            "email": "harry@potter.com",
            "role": "father",
        }

        @request.after
        def remember(response):
            identity = Identity(userid, **updated_extra_claims)
            request.app.remember_identity(response, request, identity)

        return {
            "userid": "user",
            "fullname": "Harry Potter",
            "email": "harry@potter.com",
            "role": "father",
        }

    refresh_nonce_handler = "more.jwtauth.tests.handler.refresh_nonce_handler"
    refresh_delta = 3600
    settings = {
        "master_secret": "secret",
        "allow_refresh": True,
        "refresh_delta": refresh_delta,
        "refresh_nonce_handler": refresh_nonce_handler,
    }

    @App.setting_section(section="jwtauth")
    def get_jwtauth_settings():
        return settings

    identity_policy = JWTIdentityPolicy(**settings)

    now = timegm(datetime.utcnow().utctimetuple())

    claims_set = {
        "sub": "user",
        "refresh_until": now + refresh_delta,
        "nonce": "__user__",
        "fullname": "Harry Potter",
        "email": "harry@potter.com",
        "role": "wizard",
    }

    token = identity_policy.encode_jwt(claims_set)
    headers = {"Authorization": "JWT " + token}

    morepath.commit(App)
    app = App()
    c = Client(app)

    r = c.get("/refresh", headers=headers)

    assert r.json == {
        "userid": "user",
        "fullname": "Harry Potter",
        "email": "harry@potter.com",
        "role": "father",
    }

    authtype, token = r.headers["Authorization"].split(" ", 1)
    claims_set_decoded = identity_policy.decode_jwt(token)

    assert identity_policy.get_userid(claims_set_decoded) == "user"

    extra_claims = {
        "fullname": "Harry Potter",
        "email": "harry@potter.com",
        "role": "father",
    }
    assert identity_policy.get_extra_claims(claims_set_decoded) == extra_claims


def test_refresh_delta_expired():
    class App(morepath.App):
        pass

    class Refresh:
        pass

    @App.path(model=Refresh, path="refresh")
    def get_refresh():
        return Refresh()

    @App.json(model=Refresh)
    def refresh(self, request):
        verify_refresh_request(request)

    refresh_nonce_handler = "more.jwtauth.tests.handler.refresh_nonce_handler"
    refresh_delta = timedelta(seconds=-2)
    settings = {
        "master_secret": "secret",
        "allow_refresh": True,
        "refresh_delta": refresh_delta,
        "refresh_nonce_handler": refresh_nonce_handler,
    }

    @App.setting_section(section="jwtauth")
    def get_jwtauth_settings():
        return settings

    identity_policy = JWTIdentityPolicy(**settings)

    refresh_until = timegm((datetime.utcnow() + refresh_delta).utctimetuple())

    claims_set = {
        "sub": "user",
        "refresh_until": refresh_until,
        "nonce": "__user__",
    }

    token = identity_policy.encode_jwt(claims_set)
    headers = {"Authorization": "JWT " + token}

    morepath.commit(App)
    c = Client(App())

    with pytest.raises(ExpiredSignatureError) as excinfo:
        c.get("/refresh", headers=headers)
    assert "Refresh nonce has expired" in str(excinfo.value)


def test_refresh_not_allowed():
    class App(morepath.App):
        pass

    class Refresh:
        pass

    @App.path(model=Refresh, path="refresh")
    def get_refresh():
        return Refresh()

    @App.json(model=Refresh)
    def refresh(self, request):
        verify_refresh_request(request)

    refresh_nonce_handler = "more.jwtauth.tests.handler.refresh_nonce_handler"
    refresh_delta = timedelta(seconds=3600)
    settings = {
        "master_secret": "secret",
        "allow_refresh": False,
        "refresh_delta": refresh_delta,
        "refresh_nonce_handler": refresh_nonce_handler,
    }

    @App.setting_section(section="jwtauth")
    def get_jwtauth_settings():
        return settings

    identity_policy = JWTIdentityPolicy(**settings)

    refresh_until = timegm((datetime.utcnow() + refresh_delta).utctimetuple())

    claims_set = {
        "sub": "user",
        "refresh_until": refresh_until,
        "nonce": "__user__",
    }

    token = identity_policy.encode_jwt(claims_set)
    headers = {"Authorization": "JWT " + token}

    morepath.commit(App)
    c = Client(App())

    with pytest.raises(InvalidTokenError) as excinfo:
        c.get("/refresh", headers=headers)
    assert "Token refresh is disabled" in str(excinfo.value)


def test_refresh_delta_expired_but_with_leeway():
    class App(morepath.App):
        pass

    class Refresh:
        pass

    @App.identity_policy()
    def get_identity_policy(settings):
        jwtauth_settings = settings.jwtauth.__dict__.copy()
        return JWTIdentityPolicy(**jwtauth_settings)

    @App.path(model=Refresh, path="refresh")
    def get_refresh():
        return Refresh()

    @App.json(model=Refresh)
    def refresh(self, request):
        userid = verify_refresh_request(request)

        @request.after
        def remember(response):
            identity = Identity(userid)
            request.app.remember_identity(response, request, identity)

        return {"userid": userid}

    refresh_nonce_handler = "more.jwtauth.tests.handler.refresh_nonce_handler"
    settings = {
        "master_secret": "secret",
        "leeway": timedelta(seconds=3),
        "allow_refresh": True,
        "refresh_delta": -2,
        "refresh_nonce_handler": refresh_nonce_handler,
    }

    @App.setting_section(section="jwtauth")
    def get_jwtauth_settings():
        return settings

    identity_policy = JWTIdentityPolicy(**settings)

    now = timegm(datetime.utcnow().utctimetuple())

    claims_set = {"sub": "user", "refresh_until": now - 2, "nonce": "__user__"}

    token = identity_policy.encode_jwt(claims_set)
    headers = {"Authorization": "JWT " + token}

    morepath.commit(App)
    c = Client(App())

    r = c.get("/refresh", headers=headers)

    assert r.json == {
        "userid": "user",
    }

    authtype, token = r.headers["Authorization"].split(" ", 1)
    claims_set_decoded = identity_policy.decode_jwt(token)

    assert identity_policy.get_userid(claims_set_decoded) == "user"


def test_expiration_delta_expired_with_verify_expiration_on_refresh():
    class App(morepath.App):
        pass

    class Refresh:
        pass

    @App.path(model=Refresh, path="refresh")
    def get_refresh():
        return Refresh()

    @App.json(model=Refresh)
    def refresh(self, request):
        verify_refresh_request(request)

    refresh_nonce_handler = "more.jwtauth.tests.handler.refresh_nonce_handler"
    expiration_delta = timedelta(seconds=-3)
    refresh_delta = timedelta(seconds=3)
    settings = {
        "master_secret": "secret",
        "expiration_delta": expiration_delta,
        "allow_refresh": True,
        "refresh_delta": refresh_delta,
        "refresh_nonce_handler": refresh_nonce_handler,
        "verify_expiration_on_refresh": True,
    }

    @App.setting_section(section="jwtauth")
    def get_jwtauth_settings():
        return settings

    identity_policy = JWTIdentityPolicy(**settings)

    exp = timegm((datetime.utcnow() + expiration_delta).utctimetuple())
    refresh_until = timegm((datetime.utcnow() + refresh_delta).utctimetuple())

    claims_set = {
        "sub": "user",
        "exp": exp,
        "refresh_until": refresh_until,
        "nonce": "__user__",
    }

    token = identity_policy.encode_jwt(claims_set)
    headers = {"Authorization": "JWT " + token}

    morepath.commit(App)
    c = Client(App())

    with pytest.raises(ExpiredSignatureError) as excinfo:
        c.get("/refresh", headers=headers)
    assert "Token has expired" in str(excinfo.value)


def test_expiration_delta_expired_without_verify_expiration_on_refresh():
    class App(morepath.App):
        pass

    class Refresh:
        pass

    @App.identity_policy()
    def get_identity_policy(settings):
        jwtauth_settings = settings.jwtauth.__dict__.copy()
        return JWTIdentityPolicy(**jwtauth_settings)

    @App.path(model=Refresh, path="refresh")
    def get_refresh():
        return Refresh()

    @App.json(model=Refresh)
    def refresh(self, request):
        userid = verify_refresh_request(request)

        @request.after
        def remember(response):
            identity = Identity(userid)
            request.app.remember_identity(response, request, identity)

        return {"userid": userid}

    refresh_nonce_handler = "more.jwtauth.tests.handler.refresh_nonce_handler"
    expiration_delta = -3
    refresh_delta = 3
    settings = {
        "master_secret": "secret",
        "expiration_delta": expiration_delta,
        "leeway": timedelta(seconds=3),
        "allow_refresh": True,
        "refresh_delta": refresh_delta,
        "refresh_nonce_handler": refresh_nonce_handler,
    }

    @App.setting_section(section="jwtauth")
    def get_jwtauth_settings():
        return settings

    identity_policy = JWTIdentityPolicy(**settings)

    now = timegm(datetime.utcnow().utctimetuple())

    claims_set = {
        "sub": "user",
        "exp": now + expiration_delta,
        "refresh_until": now + refresh_delta,
        "nonce": "__user__",
    }

    token = identity_policy.encode_jwt(claims_set)
    headers = {"Authorization": "JWT " + token}

    morepath.commit(App)
    c = Client(App())

    r = c.get("/refresh", headers=headers)

    assert r.json == {
        "userid": "user",
    }

    authtype, token = r.headers["Authorization"].split(" ", 1)
    claims_set_decoded = identity_policy.decode_jwt(token)

    assert identity_policy.get_userid(claims_set_decoded) == "user"


def test_refresh_without_refresh_nonce_handler_setting():
    class App(morepath.App):
        pass

    class Refresh:
        pass

    @App.identity_policy()
    def get_identity_policy(settings):
        jwtauth_settings = settings.jwtauth.__dict__.copy()
        return JWTIdentityPolicy(**jwtauth_settings)

    @App.path(model=Refresh, path="refresh")
    def get_refresh():
        return Refresh()

    @App.json(model=Refresh)
    def refresh(self, request):
        userid = verify_refresh_request(request)

        @request.after
        def remember(response):
            identity = Identity(userid)
            request.app.remember_identity(response, request, identity)

        return {"userid": userid}

    settings = {
        "master_secret": "secret",
        "allow_refresh": True,
        "refresh_delta": 3,
    }

    @App.setting_section(section="jwtauth")
    def get_jwtauth_settings():
        return settings

    identity_policy = JWTIdentityPolicy(**settings)

    now = timegm(datetime.utcnow().utctimetuple())

    claims_set = {
        "sub": "user",
        "refresh_until": now + 3,
    }

    token = identity_policy.encode_jwt(claims_set)
    headers = {"Authorization": "JWT " + token}

    morepath.commit(App)
    c = Client(App())

    r = c.get("/refresh", headers=headers)

    assert r.json == {
        "userid": "user",
    }

    authtype, token = r.headers["Authorization"].split(" ", 1)
    claims_set_decoded = identity_policy.decode_jwt(token)

    assert identity_policy.get_userid(claims_set_decoded) == "user"


def test_refresh_without_token():
    class App(morepath.App):
        pass

    class Refresh:
        pass

    @App.path(model=Refresh, path="refresh")
    def get_refresh():
        return Refresh()

    @App.json(model=Refresh)
    def refresh(self, request):
        verify_refresh_request(request)

    refresh_nonce_handler = "more.jwtauth.tests.handler.refresh_nonce_handler"

    @App.setting_section(section="jwtauth")
    def get_jwtauth_settings():
        return {
            "master_secret": "secret",
            "allow_refresh": True,
            "refresh_delta": 3,
            "refresh_nonce_handler": refresh_nonce_handler,
        }

    morepath.commit(App)
    c = Client(App())

    with pytest.raises(InvalidTokenError) as excinfo:
        c.get("/refresh")
    assert "Token not found" in str(excinfo.value)


def test_refresh_with_invalid_token():
    class App(morepath.App):
        pass

    class Refresh:
        pass

    @App.path(model=Refresh, path="refresh")
    def get_refresh():
        return Refresh()

    @App.json(model=Refresh)
    def refresh(self, request):
        verify_refresh_request(request)

    refresh_nonce_handler = "more.jwtauth.tests.handler.refresh_nonce_handler"

    @App.setting_section(section="jwtauth")
    def get_jwtauth_settings():
        return {
            "master_secret": "secret",
            "allow_refresh": True,
            "refresh_delta": 3,
            "refresh_nonce_handler": refresh_nonce_handler,
        }

    token = "Invalid Token"
    headers = {"Authorization": "JWT " + token}

    morepath.commit(App)
    c = Client(App())

    with pytest.raises(DecodeError) as excinfo:
        c.get("/refresh", headers=headers)
    assert "Token could not be decoded" in str(excinfo.value)


def test_refresh_with_invalid_refresh_nonce():
    class App(morepath.App):
        pass

    class Refresh:
        pass

    @App.path(model=Refresh, path="refresh")
    def get_refresh():
        return Refresh()

    @App.json(model=Refresh)
    def refresh(self, request):
        verify_refresh_request(request)

    refresh_nonce_handler = "more.jwtauth.tests.handler.refresh_nonce_handler"
    settings = {
        "master_secret": "secret",
        "allow_refresh": True,
        "refresh_delta": 3,
        "refresh_nonce_handler": refresh_nonce_handler,
    }

    @App.setting_section(section="jwtauth")
    def get_jwtauth_settings():
        return settings

    identity_policy = JWTIdentityPolicy(**settings)

    now = timegm(datetime.utcnow().utctimetuple())

    claims_set = {
        "sub": "user",
        "refresh_until": now + 3,
        "nonce": "__invalid__",
    }

    token = identity_policy.encode_jwt(claims_set)
    headers = {"Authorization": "JWT " + token}

    morepath.commit(App)
    c = Client(App())

    with pytest.raises(InvalidTokenError) as excinfo:
        c.get("/refresh", headers=headers)
    assert "Refresh nonce is not valid" in str(excinfo.value)


def test_refresh_with_missing_userid_claim():
    class App(morepath.App):
        pass

    class Refresh:
        pass

    @App.path(model=Refresh, path="refresh")
    def get_refresh():
        return Refresh()

    @App.json(model=Refresh)
    def refresh(self, request):
        verify_refresh_request(request)

    refresh_nonce_handler = "more.jwtauth.tests.handler.refresh_nonce_handler"
    settings = {
        "master_secret": "secret",
        "allow_refresh": True,
        "refresh_delta": 3,
        "refresh_nonce_handler": refresh_nonce_handler,
    }

    @App.setting_section(section="jwtauth")
    def get_jwtauth_settings():
        return settings

    identity_policy = JWTIdentityPolicy(**settings)

    now = timegm(datetime.utcnow().utctimetuple())

    claims_set = {"refresh_until": now + 3, "nonce": "__user__"}

    token = identity_policy.encode_jwt(claims_set)
    headers = {"Authorization": "JWT " + token}

    morepath.commit(App)
    c = Client(App())

    with pytest.raises(MissingRequiredClaimError) as excinfo:
        c.get("/refresh", headers=headers)
    assert 'Token is missing the "sub" claim' in str(excinfo.value)


def test_refresh_with_missing_refresh_until_claim():
    class App(morepath.App):
        pass

    class Refresh:
        pass

    @App.path(model=Refresh, path="refresh")
    def get_refresh():
        return Refresh()

    @App.json(model=Refresh)
    def refresh(self, request):
        verify_refresh_request(request)

    refresh_nonce_handler = "more.jwtauth.tests.handler.refresh_nonce_handler"
    settings = {
        "master_secret": "secret",
        "allow_refresh": True,
        "refresh_delta": 3,
        "refresh_nonce_handler": refresh_nonce_handler,
    }

    @App.setting_section(section="jwtauth")
    def get_jwtauth_settings():
        return settings

    identity_policy = JWTIdentityPolicy(**settings)

    claims_set = {"sub": "user", "nonce": "__user__"}

    token = identity_policy.encode_jwt(claims_set)
    headers = {"Authorization": "JWT " + token}

    morepath.commit(App)
    c = Client(App())

    with pytest.raises(MissingRequiredClaimError) as excinfo:
        c.get("/refresh", headers=headers)
    assert 'Token is missing the "refresh_until" claim' in str(excinfo.value)


def test_refresh_with_missing_nonce_claim():
    class App(morepath.App):
        pass

    class Refresh:
        pass

    @App.path(model=Refresh, path="refresh")
    def get_refresh():
        return Refresh()

    @App.json(model=Refresh)
    def refresh(self, request):
        verify_refresh_request(request)

    refresh_nonce_handler = "more.jwtauth.tests.handler.refresh_nonce_handler"
    settings = {
        "master_secret": "secret",
        "allow_refresh": True,
        "refresh_delta": 3,
        "refresh_nonce_handler": refresh_nonce_handler,
    }

    @App.setting_section(section="jwtauth")
    def get_jwtauth_settings():
        return settings

    identity_policy = JWTIdentityPolicy(**settings)

    now = timegm(datetime.utcnow().utctimetuple())

    claims_set = {"sub": "user", "refresh_until": now + 3}

    token = identity_policy.encode_jwt(claims_set)
    headers = {"Authorization": "JWT " + token}

    morepath.commit(App)
    c = Client(App())

    with pytest.raises(MissingRequiredClaimError) as excinfo:
        c.get("/refresh", headers=headers)
    assert 'Token is missing the "nonce" claim' in str(excinfo.value)



from importlib import import_module


def import_dotted_path(path):
    """
    Takes a dotted path to a member name in a module, and returns
    the member after importing it.
    """
    try:
        module_path, member_name = path.rsplit(".", 1)
        module = import_module(module_path)
        return getattr(module, member_name)
    except (ValueError, ImportError, AttributeError) as e:
        raise ImportError("Could not import the name: {}: {}".format(path, e))


def handler(path):
    """
    Returns a handler from a dotted path and None if no path given.
    """
    if path:
        return import_dotted_path(path)
    else:
        return None



# flake8: noqa

from jwt import (
    InvalidTokenError,
    DecodeError,
    ExpiredSignatureError,
    MissingRequiredClaimError,
    InvalidIssuerError,
)

from .main import JWTIdentityPolicy
from .refresh import verify_refresh_request



import pytest

from more.jwtauth import JWTIdentityPolicy

try:
    from cryptography.hazmat.primitives.asymmetric import ec  # noqa

    has_crypto = True
except ImportError:  # pragma: no cover
    has_crypto = False


def relative(filepath):
    import os

    return os.path.join(os.path.dirname(__file__), filepath)


@pytest.mark.skipif(
    not has_crypto, reason="Not supported without cryptography library"
)
class TestCryptographyAlgorithms:
    def test_encode_decode_with_es256_as_bytes(self):
        with open(relative("keys/testkey_ec")) as key_priv_file:
            private_key = key_priv_file.read()
        with open(relative("keys/testkey_ec.pub")) as key_pub_file:
            public_key = key_pub_file.read()

        identity_policy = JWTIdentityPolicy(
            algorithm="ES256", private_key=private_key, public_key=public_key
        )
        claims_set = {"sub": "user"}
        token = identity_policy.encode_jwt(claims_set)
        claims_set_decoded = identity_policy.decode_jwt(token)

        assert claims_set_decoded == claims_set

    def test_encode_decode_with_es256(self):
        identity_policy = JWTIdentityPolicy(
            algorithm="ES256",
            private_key_file=relative("keys/testkey_ec"),
            public_key_file=relative("keys/testkey_ec.pub"),
        )
        claims_set = {"sub": "user"}
        token = identity_policy.encode_jwt(claims_set)
        claims_set_decoded = identity_policy.decode_jwt(token)

        assert claims_set_decoded == claims_set

    def test_encode_decode_with_es384(self):
        identity_policy = JWTIdentityPolicy(
            algorithm="ES384",
            private_key_file=relative("keys/testkey_ec"),
            public_key_file=relative("keys/testkey_ec.pub"),
        )
        claims_set = {"sub": "user"}
        token = identity_policy.encode_jwt(claims_set)
        claims_set_decoded = identity_policy.decode_jwt(token)

        assert claims_set_decoded == claims_set

    def test_encode_decode_with_es512(self):
        identity_policy = JWTIdentityPolicy(
            algorithm="ES512",
            private_key_file=relative("keys/testkey_ec"),
            public_key_file=relative("keys/testkey_ec.pub"),
        )
        claims_set = {"sub": "user"}
        token = identity_policy.encode_jwt(claims_set)
        claims_set_decoded = identity_policy.decode_jwt(token)

        assert claims_set_decoded == claims_set

    def test_encode_decode_with_ps256(self):
        identity_policy = JWTIdentityPolicy(
            algorithm="PS256",
            private_key_file=relative("keys/testkey_rsa"),
            public_key_file=relative("keys/testkey_rsa.pub"),
        )
        claims_set = {"sub": "user"}
        token = identity_policy.encode_jwt(claims_set)
        claims_set_decoded = identity_policy.decode_jwt(token)

        assert claims_set_decoded == claims_set

    def test_encode_decode_with_ps384(self):
        identity_policy = JWTIdentityPolicy(
            algorithm="PS384",
            private_key_file=relative("keys/testkey_rsa"),
            public_key_file=relative("keys/testkey_rsa.pub"),
        )
        claims_set = {"sub": "user"}
        token = identity_policy.encode_jwt(claims_set)
        claims_set_decoded = identity_policy.decode_jwt(token)

        assert claims_set_decoded == claims_set

    def test_encode_decode_with_ps512(self):
        identity_policy = JWTIdentityPolicy(
            algorithm="PS512",
            private_key_file=relative("keys/testkey_rsa"),
            public_key_file=relative("keys/testkey_rsa.pub"),
        )
        claims_set = {"sub": "user"}
        token = identity_policy.encode_jwt(claims_set)
        claims_set_decoded = identity_policy.decode_jwt(token)

        assert claims_set_decoded == claims_set

    def test_encode_decode_with_rs256(self):
        identity_policy = JWTIdentityPolicy(
            algorithm="RS256",
            private_key_file=relative("keys/testkey_rsa"),
            public_key_file=relative("keys/testkey_rsa.pub"),
        )
        claims_set = {"sub": "user"}
        token = identity_policy.encode_jwt(claims_set)
        claims_set_decoded = identity_policy.decode_jwt(token)

        assert claims_set_decoded == claims_set

    def test_encode_decode_with_rs384(self):
        identity_policy = JWTIdentityPolicy(
            algorithm="RS384",
            private_key_file=relative("keys/testkey_rsa"),
            public_key_file=relative("keys/testkey_rsa.pub"),
        )
        claims_set = {"sub": "user"}
        token = identity_policy.encode_jwt(claims_set)
        claims_set_decoded = identity_policy.decode_jwt(token)

        assert claims_set_decoded == claims_set

    def test_encode_decode_with_rs512(self):
        identity_policy = JWTIdentityPolicy(
            algorithm="RS512",
            private_key_file=relative("keys/testkey_rsa"),
            public_key_file=relative("keys/testkey_rsa.pub"),
        )
        claims_set = {"sub": "user"}
        token = identity_policy.encode_jwt(claims_set)
        claims_set_decoded = identity_policy.decode_jwt(token)

        assert claims_set_decoded == claims_set



from datetime import timedelta

import pytest

import morepath
from more.jwtauth import (
    JWTIdentityPolicy,
    ExpiredSignatureError,
    InvalidIssuerError,
)
from morepath import Identity, NO_IDENTITY, Response
from webob.exc import HTTPProxyAuthenticationRequired
from webtest import TestApp as Client


def test_jwt_custom_settings():
    class App(morepath.App):
        pass

    @App.setting_section(section="jwtauth")
    def get_jwtauth_settings():
        return {
            "public_key": "MIGbMBAGByqGSM49AgEGBSuBBAAjA4GGAAQBWcJwPEAnS/k4kFg"
            "UhxNF7J0SQQhZG+nNgy+/mXwhQ5PZIUmId1a1TjkNXiKzv6Dptt"
            "BqduHbz/V0EtH+QfWy0B4BhZ5MnTyDGjcz1DQqKdexebhzobbhS"
            "IZjpYd5aU48o9rXp/OnAnrajddpGsJ0bNf4rtMLBqFYJN6LOslA"
            "B7xTBRg=",
            "algorithm": "ES256",
            "leeway": 20,
        }

    app = App()
    app.commit()

    assert app.settings.jwtauth.algorithm == "ES256"
    assert app.settings.jwtauth.leeway == 20
    assert (
        app.settings.jwtauth.public_key
        == "MIGbMBAGByqGSM49AgEGBSuBBAAjA4GGAAQBWcJwPEAnS/k4kFgUhxNF7J0SQQhZG+nN"
        "gy+/mXwhQ5PZIUmId1a1TjkNXiKzv6DpttBqduHbz/V0EtH+QfWy0B4BhZ5MnTyDGjcz"
        "1DQqKdexebhzobbhSIZjpYd5aU48o9rXp/OnAnrajddpGsJ0bNf4rtMLBqFYJN6LOslA"
        "B7xTBRg="
    )


def test_encode_decode():
    identity_policy = JWTIdentityPolicy(master_secret="secret")
    claims_set = {"sub": "user"}
    token = identity_policy.encode_jwt(claims_set)
    claims_set_decoded = identity_policy.decode_jwt(token)

    assert claims_set_decoded == claims_set


def test_encode_decode_with_unicode():
    identity_policy = JWTIdentityPolicy(master_secret="sëcret")
    claims_set = {"sub": "user"}
    token = identity_policy.encode_jwt(claims_set)
    claims_set_decoded = identity_policy.decode_jwt(token)

    assert claims_set_decoded == claims_set


def test_encode_decode_with_issuer():
    identity_policy = JWTIdentityPolicy(
        master_secret="secret", issuer="Issuer_App"
    )
    userid = "user"

    extra_claims = {"iss": "Invalid_Issuer_App"}
    claims_set = identity_policy.create_claims_set(None, userid, extra_claims)
    token = identity_policy.encode_jwt(claims_set)

    with pytest.raises(InvalidIssuerError):
        claims_set_decoded = identity_policy.decode_jwt(token)

    extra_claims = {"iss": "Issuer_App"}
    claims_set = identity_policy.create_claims_set(None, userid, extra_claims)
    token = identity_policy.encode_jwt(claims_set)
    claims_set_decoded = identity_policy.decode_jwt(token)

    assert claims_set_decoded == claims_set


def test_create_claims_and_encode_decode_and_get_userid_and_get_extra_claims():
    identity_policy = JWTIdentityPolicy(master_secret="secret")
    userid = "user"
    extra_claims = {"email": "user@example.com", "role": "admin"}
    claims_set = identity_policy.create_claims_set(None, userid, extra_claims)
    token = identity_policy.encode_jwt(claims_set)
    claims_set_decoded = identity_policy.decode_jwt(token)

    assert userid == identity_policy.get_userid(claims_set_decoded)
    assert extra_claims == identity_policy.get_extra_claims(claims_set_decoded)


def test_get_userid_without_userid():
    identity_policy = JWTIdentityPolicy(master_secret="secret")
    claims_set = {}
    token = identity_policy.encode_jwt(claims_set)
    claims_set_decoded = identity_policy.decode_jwt(token)

    assert identity_policy.get_userid(claims_set_decoded) is None


def test_create_claims_and_encode_decode_expired():
    identity_policy = JWTIdentityPolicy(
        master_secret="secret", expiration_delta=timedelta(seconds=-2)
    )
    userid = "user"
    claims_set = identity_policy.create_claims_set(None, userid)
    token = identity_policy.encode_jwt(claims_set)

    with pytest.raises(ExpiredSignatureError) as excinfo:
        identity_policy.decode_jwt(token)
    assert "Signature has expired" in str(excinfo.value)


def test_create_claims_and_encode_decode_expired_but_with_leeway():
    identity_policy = JWTIdentityPolicy(
        master_secret="secret", expiration_delta=-2, leeway=3
    )
    userid = "user"
    claims_set = identity_policy.create_claims_set(None, userid)
    token = identity_policy.encode_jwt(claims_set)
    claims_set_decoded = identity_policy.decode_jwt(token)

    assert identity_policy.get_userid(claims_set_decoded) == userid


def test_login():
    class App(morepath.App):
        pass

    @App.identity_policy()
    def get_identity_policy(settings):
        jwtauth_settings = settings.jwtauth.__dict__.copy()
        return JWTIdentityPolicy(**jwtauth_settings)

    class Login:
        pass

    @App.path(model=Login, path="login")
    def get_login():
        return Login()

    def user_has_password(username, password):
        return username == "user" and password == "password"

    @App.json(model=Login, request_method="POST")
    def login(self, request):
        username = request.POST["username"]
        password = request.POST["password"]
        if not user_has_password(username, password):
            raise HTTPProxyAuthenticationRequired("Invalid username/password")

        @request.after
        def remember(response):
            identity = Identity(username)
            request.app.remember_identity(response, request, identity)

        return {
            "username": username,
        }

    settings = {
        "master_secret": "secret",
        "expiration_delta": None,
        "refresh_delta": None,
    }

    @App.setting_section(section="jwtauth")
    def get_jwtauth_settings():
        return settings

    identity_policy = JWTIdentityPolicy(**settings)

    morepath.commit(App)
    c = Client(App())
    r = c.post("/login", "username=user&password=false", status=407)
    r = c.post("/login", "username=not_exists&password=password", status=407)
    r = c.post("/login", "username=user&password=password")

    assert r.json == {
        "username": "user",
    }

    claims_set = {"sub": "user"}
    expected_token = identity_policy.encode_jwt(claims_set)
    assert r.headers["Authorization"] == "{} {}".format("JWT", expected_token)

    authtype, token = r.headers["Authorization"].split(" ", 1)
    claims_set_decoded = identity_policy.decode_jwt(token)

    assert identity_policy.get_userid(claims_set_decoded) == "user"


def test_login_with_extra_claims():
    class App(morepath.App):
        pass

    @App.identity_policy()
    def get_identity_policy(settings):
        jwtauth_settings = settings.jwtauth.__dict__.copy()
        return JWTIdentityPolicy(**jwtauth_settings)

    class Login:
        pass

    @App.path(model=Login, path="login")
    def get_login():
        return Login()

    def user_has_password(username, password):
        return username == "user" and password == "password"

    @App.json(model=Login, request_method="POST")
    def login(self, request):
        username = request.POST["username"]
        password = request.POST["password"]
        fullname = request.POST["fullname"]
        email = request.POST["email"]
        role = request.POST["role"]

        if not user_has_password(username, password):
            raise HTTPProxyAuthenticationRequired("Invalid username/password")

        @request.after
        def remember(response):
            identity = Identity(
                username, fullname=fullname, email=email, role=role
            )
            request.app.remember_identity(response, request, identity)

        return {
            "username": username,
            "fullname": fullname,
            "email": email,
            "role": role,
        }

    settings = {"master_secret": "secret"}

    @App.setting_section(section="jwtauth")
    def get_jwtauth_settings():
        return settings

    identity_policy = JWTIdentityPolicy(**settings)

    morepath.commit(App)
    c = Client(App())
    params = {
        "username": "not_exist",
        "password": "password",
        "fullname": "Harry Potter",
        "email": "harry@potter.com",
        "role": "wizard",
    }
    r = c.post("/login", params, status=407)
    params = {
        "username": "user",
        "password": "password",
        "fullname": "Harry Potter",
        "email": "harry@potter.com",
        "role": "wizard",
    }
    r = c.post("/login", params)

    assert r.json == {
        "username": "user",
        "fullname": "Harry Potter",
        "email": "harry@potter.com",
        "role": "wizard",
    }

    authtype, token = r.headers["Authorization"].split(" ", 1)
    claims_set_decoded = identity_policy.decode_jwt(token)

    assert identity_policy.get_userid(claims_set_decoded) == "user"

    extra_claims = {
        "fullname": "Harry Potter",
        "email": "harry@potter.com",
        "role": "wizard",
    }
    assert identity_policy.get_extra_claims(claims_set_decoded) == extra_claims


def test_jwt_identity_policy():
    class App(morepath.App):
        pass

    class Model:
        def __init__(self, id):
            self.id = id

    class Permission:
        pass

    @App.path(
        model=Model, path="{id}", variables=lambda model: {"id": model.id}
    )
    def get_model(id):
        return Model(id)

    @App.permission_rule(model=Model, permission=Permission)
    def get_permission(identity, model, permission):
        return identity.userid == "user"

    @App.view(model=Model, permission=Permission)
    def default(self, request):
        return "Model: %s" % self.id

    @App.identity_policy()
    def get_identity_policy(settings):
        jwtauth_settings = settings.jwtauth.__dict__.copy()
        return JWTIdentityPolicy(**jwtauth_settings)

    @App.verify_identity()
    def verify_identity(identity):
        assert identity is not NO_IDENTITY
        return True

    settings = {
        "master_secret": "secret",
    }

    @App.setting_section(section="jwtauth")
    def get_jwtauth_settings():
        return settings

    identity_policy = JWTIdentityPolicy(**settings)

    morepath.commit(App)

    c = Client(App())

    response = c.get("/foo", status=403)

    claims_set = {"sub": "wrong_user"}
    token = identity_policy.encode_jwt(claims_set)
    headers = {"Authorization": "JWT " + token}
    response = c.get("/foo", headers=headers, status=403)

    claims_set = {"sub": "user"}
    token = identity_policy.encode_jwt(claims_set)
    headers = {"Authorization": "JWT " + token}
    response = c.get("/foo", headers=headers)
    assert response.body == b"Model: foo"

    # extra claims
    claims_set = {
        "sub": "user",
        "email": "harry@potter.com",
    }
    token = identity_policy.encode_jwt(claims_set)
    headers = {"Authorization": "JWT " + token}
    response = c.get("/foo", headers=headers)
    assert response.body == b"Model: foo"


def test_jwt_identity_policy_errors():
    class App(morepath.App):
        pass

    class Model:
        def __init__(self, id):
            self.id = id

    class Permission:
        pass

    @App.path(
        model=Model, path="{id}", variables=lambda model: {"id": model.id}
    )
    def get_model(id):
        return Model(id)

    @App.permission_rule(model=Model, permission=Permission)
    def get_permission(identity, model, permission):
        return identity.userid == "user"

    @App.view(model=Model, permission=Permission)
    def default(self, request):
        return "Model: %s" % self.id

    settings = {"master_secret": "sëcret", "leeway": None}

    @App.setting_section(section="jwtauth")
    def get_jwtauth_settings():
        return settings

    @App.identity_policy()
    def get_identity_policy(settings):
        jwtauth_settings = settings.jwtauth.__dict__.copy()
        return JWTIdentityPolicy(**jwtauth_settings)

    @App.verify_identity()
    def verify_identity(identity):
        assert identity is not NO_IDENTITY
        return True

    identity_policy = JWTIdentityPolicy(**settings)

    morepath.commit(App)

    c = Client(App())

    response = c.get("/foo", status=403)

    headers = {"Authorization": "Something"}
    response = c.get("/foo", headers=headers, status=403)

    headers = {"Authorization": "JWT " + "nonsense"}
    response = c.get("/foo", headers=headers, status=403)

    # no userid
    claims_set = {"sub": None}
    token = identity_policy.encode_jwt(claims_set)
    headers = {"Authorization": "JWT " + token}
    response = c.get("/foo", headers=headers, status=403)

    claims_set = {"sub": "user"}
    token = identity_policy.encode_jwt(claims_set)

    headers = {"Authorization": "OtherAuthType " + token}
    response = c.get("/foo", headers=headers, status=403)

    headers = {"Authorization": "JWT " + token}
    response = c.get("/foo", headers=headers)
    assert response.body == b"Model: foo"


def test_jwt_remember():
    class App(morepath.App):
        pass

    @App.path(path="{id}", variables=lambda model: {"id": model.id})
    class Model:
        def __init__(self, id):
            self.id = id

    @App.view(model=Model)
    def default(self, request):
        response = Response()
        request.app.remember_identity(response, request, Identity("foo"))
        return response

    settings = {
        "master_secret": "secret",
    }

    @App.setting_section(section="jwtauth")
    def get_jwtauth_settings():
        return settings

    @App.identity_policy()
    def get_identity_policy(settings):
        jwtauth_settings = settings.jwtauth.__dict__.copy()
        return JWTIdentityPolicy(**jwtauth_settings)

    morepath.commit(App)

    c = Client(App())

    response = c.get("/foo", status=200)
    assert response.body == b""
    assert isinstance(response.headers["Authorization"], str)
    assert response.headers["Authorization"][:7] == "JWT eyJ"


def test_jwt_forget():
    class App(morepath.App):
        pass

    @App.path(path="{id}")
    class Model:
        def __init__(self, id):
            self.id = id

    @App.view(model=Model)
    def default(self, request):
        # will not actually do anything as it's a no-op for JWT
        # auth, but at least won't crash
        response = Response(content_type="text/plain")
        request.app.forget_identity(response, request)
        return response

    @App.identity_policy()
    def get_identity_policy():
        return JWTIdentityPolicy()

    morepath.commit(App)

    c = Client(App())

    response = c.get("/foo", status=200)
    assert response.body == b""



from setuptools import find_packages, setup

long_description = (
    open("README.rst", encoding="utf-8").read()
    + "\n\n"
    + open("CHANGES.rst", encoding="utf-8").read()
)

setup(
    name="more.jwtauth",
    version="0.12",
    description="JWT Access Auth Identity Policy for Morepath",
    long_description=long_description,
    author="Morepath developers",
    author_email="morepath@googlegroups.com",
    keywords="morepath JWT identity authentication",
    license="BSD",
    url="https://github.com/morepath/more.jwtauth",
    namespace_packages=["more"],
    packages=find_packages(),
    include_package_data=True,
    zip_safe=False,
    classifiers=[
        "Intended Audience :: Developers",
        "License :: OSI Approved :: BSD License",
        "Environment :: Web Environment",
        "Topic :: Software Development :: Libraries :: Application Frameworks",
        "Topic :: Internet :: WWW/HTTP :: WSGI",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: Implementation :: PyPy",
        "Development Status :: 5 - Production/Stable",
    ],
    install_requires=["morepath >= 0.19", "PyJWT == 1.7.1"],
    extras_require=dict(
        crypto=["cryptography == 2.9.2"],
        test=[
            "pytest >= 2.9.1",
            "pytest-remove-stale-bytecode",
            "WebTest >= 2.0.14",
        ],
        pep8=["flake8", "black"],
        coverage=["pytest-cov"],
    ),
)



"""Provides the JWTIdentityPolicy.

The following settings are available:

    * master_secret:  A secret known only by the server, used for
        the default HMAC (HS*) algorithm.

    * private_key:  An Elliptic Curve or an RSA private_key used for
        the EC (EC*) or RSA (PS*/RS*) algorithms.
    * private_key_file: A file holding an Elliptic Curve or an RSA encoded
        (PEM/DER) private_key.

    * public_key:  An Elliptic Curve or an RSA public_key used for the EC (EC*)
        or RSA (PS*/RS*) algorithms.
    * public_key_file: A file holding an Elliptic Curve
        or an RSA encoded (PEM/DER) public_key.

    * algorithm:  The algorithm used to sign the key (defaults to HS256).

    * expiration_delta: Time delta from now until the token will expire.
        This can either be a datetime.timedelta or the number of seconds.
        Default is 30 minutes, set to None to disable expiration.

    * leeway:  The leeway, which allows you to validate an expiration time
        which is in the past, but not very far. To use as a datetime.timedelta
        or the number of seconds. Defaults is 0.

    * allow_refresh: Enables the token refresh API when True.
        Default is False

    * refresh_delta: A time delta in which the token can be refreshed
        considering the leeway.
        This can either be a datetime.timedelta or the number of seconds.
        Default is 7 days. When None you can always refresh the token.

    * refresh_nonce_handler: Either dotted path to callback function or the
        callback function itself, which receives the userid as argument and
        returns a nonce which will be validated before refreshing.
        When None no nonce will be created or validated for refreshing.
        Default is None.

    * verify_expiration_on_refresh: If False, expiration_delta for the JWT
        token will not be checked during refresh. Otherwise you can refresh
        the token only if it's not yet expired. Default is False.

    * issuer: This is a string that will be checked against the iss claim of
        the token. You can use this e.g. if you have several related apps with
        exclusive user audience. Default is None (do not check iss on JWT).

    * auth_header_prefix: You can modify the Authorization header value prefix
        that is required to be sent together with the token. The default value
        is JWT. Another common value used for tokens is Bearer.

    * userid_claim: The claim, which contains the user id.
        The default claim is 'sub'.

The library takes either a master_secret or private_key/public_key pair.
In the later case the algorithm must be an EC*, PS* or RS* version.
"""


from calendar import timegm
from datetime import datetime, timedelta

import jwt
from morepath import Identity, NO_IDENTITY

from . import (
    InvalidTokenError,
    DecodeError,
    ExpiredSignatureError,
    MissingRequiredClaimError,
)
from .utils import handler


class JWTIdentityPolicy:
    """Morepath Identity Policy implementing JWT Access Auth.

    This class provides an IdentityPolicy implementation based on
    signed requests, using the JSON Web Token Authentication standard.

    Reference:
    http://self-issued.info/docs/draft-ietf-oauth-json-web-token.html
    """

    def __init__(
        self,
        master_secret=None,
        private_key=None,
        private_key_file=None,
        public_key=None,
        public_key_file=None,
        algorithm="HS256",
        expiration_delta=timedelta(minutes=30),
        leeway=0,
        allow_refresh=False,
        refresh_delta=timedelta(days=7),
        refresh_nonce_handler=None,
        verify_expiration_on_refresh=False,
        issuer=None,
        auth_header_prefix="JWT",
        userid_claim="sub",
    ):
        """Initiate the JWTIdentityPolicy with the given settings."""
        _public_key = master_secret
        if public_key is not None:
            _public_key = public_key
        if public_key_file is not None:
            with open(public_key_file) as key_pub_file:
                _public_key = key_pub_file.read()
        self.public_key = _public_key

        _private_key = master_secret
        if private_key is not None:
            _private_key = private_key
        if private_key_file is not None:
            with open(private_key_file) as key_priv_file:
                _private_key = key_priv_file.read()
        self.private_key = _private_key

        self.algorithm = algorithm

        if isinstance(expiration_delta, timedelta):
            expiration_delta = expiration_delta.total_seconds()
        self.expiration_delta = expiration_delta

        if leeway is None:
            leeway = 0
        elif isinstance(leeway, timedelta):
            leeway = leeway.total_seconds()
        self.leeway = leeway

        self.allow_refresh = allow_refresh

        if isinstance(refresh_delta, timedelta):
            refresh_delta = refresh_delta.total_seconds()
        self.refresh_delta = refresh_delta

        if isinstance(refresh_nonce_handler, str):
            self.refresh_nonce_handler = handler(refresh_nonce_handler)
        else:
            self.refresh_nonce_handler = refresh_nonce_handler
        self.verify_expiration_on_refresh = verify_expiration_on_refresh
        self.issuer = issuer
        self.auth_header_prefix = auth_header_prefix
        self.userid_claim = userid_claim

    def identify(self, request):
        """Establish what identity this user claims to have from request.

        :param request: Request to extract identity information from.
        :type request: :class:`morepath.Request`.
        :returns: :class:`morepath.Identity` instance or
          :attr:`morepath.NO_IDENTITY` if identity cannot
          be established.
        """
        token = self.get_jwt(request)
        if token is None:
            return NO_IDENTITY
        try:
            claims_set = self.decode_jwt(token)
        except (DecodeError, ExpiredSignatureError):
            return NO_IDENTITY
        userid = self.get_userid(claims_set)
        if userid is None:
            return NO_IDENTITY
        extra_claims = self.get_extra_claims(claims_set)
        if extra_claims is not None:
            return Identity(userid=userid, **extra_claims)
        else:
            return Identity(userid=userid)

    def remember(self, response, request, identity):
        """Remember identity on response.

        Implements ``morepath.App.remember_identity``, which is called
        from user login code.

        Create a JWT token and return it as the Authorization field of the
        response header.

        :param response: response object on which to store identity.
        :type response: :class:`morepath.Response`
        :param request: request object.
        :type request: :class:`morepath.Request`
        :param identity: identity to remember.
        :type identity: :class:`morepath.Identity`
        """
        claims = identity.as_dict()
        userid = claims.pop("userid")
        claims_set = self.create_claims_set(request, userid, claims)
        token = self.encode_jwt(claims_set)
        response.headers["Authorization"] = "{} {}".format(
            self.auth_header_prefix, token,
        )

    def forget(self, response, request):
        """Forget identity on response.

        Implements ``morepath.App.forget_identity``, which is called from
        user logout code.

        This is a no-op for this identity policy. The client is supposed to
        handle logout and remove the token.

        :param response: response object on which to forget identity.
        :type response: :class:`morepath.Response`
        :param request: request object.
        :type request: :class:`morepath.Request`
        """
        pass

    def decode_jwt(self, token, verify_expiration=True):
        """Decode a JWTAuth token into its claims set.

        This method decodes the given JWT to provide the claims set.  The JWT
        can fail if the token has expired (with appropriate leeway) or if the
        token won't validate due to the secret (key) being wrong.

        If private_key/public key is set then the public_key will be used
        to decode the key.
        The leeway and issuer settings will be passed to jwt.decode.

        :param token: the JWTAuth token.
        :param verify_expiration: if False the expiration time will not
            be checked.
        """
        options = {
            "verify_exp": verify_expiration,
        }
        return jwt.decode(
            token,
            self.public_key,
            algorithms=[self.algorithm],
            options=options,
            leeway=self.leeway,
            issuer=self.issuer,
        )

    def create_claims_set(self, request, userid, extra_claims=None):
        """Create the claims set based on the userid of the claimed identity,
        the settings and the extra_claims dictionary.

        The userid will be stored in settings.jwtauth.userid_claim
        (default: "sub").
        If settings.jwtauth.expiration_delta is set it will be added
        to the current time and stored in the "exp" claim.
        If settings.jwtauth.issuer is set, it get stored in the "iss" claim.
        If settings.jwtauth.refresh_delta is set it will be added
        to the current time and stored in the "refresh_until" claim and
        the return value of settings.jwtauth.refresh_nonce_handler called with
        "user_id" as argument will be stored in the "nonce" claim.

        With the extra_claims dictionary you can provide additional claims.
        This can be registered claims like "nbf"
        (the time before which the token should not be processed) and/or
        claims containing extra info
        about the identity, which will be stored in the Identity object.

        :param request: current request object.
        :type request: :class:`morepath.Request`
        :param userid:  the userid of the claimed identity.
        :param extra_claims: dictionary, containing additional claims or None.
        """
        claims_set = {self.userid_claim: userid}
        now = timegm(datetime.utcnow().utctimetuple())
        if self.expiration_delta is not None:
            claims_set["exp"] = now + self.expiration_delta
        if self.issuer is not None:
            claims_set["iss"] = self.issuer
        if self.allow_refresh:
            if self.refresh_delta is not None:
                claims_set["refresh_until"] = now + self.refresh_delta
            if self.refresh_nonce_handler is not None:
                claims_set["nonce"] = self.refresh_nonce_handler(
                    request, userid
                )
        if extra_claims is not None:
            claims_set.update(extra_claims)
        return claims_set

    def encode_jwt(self, claims_set):
        """Encode a JWT token based on the claims_set and the settings.

        If available, registry.settings.jwtauth.private_key is used as key.
        In this case the algorithm must be an RS* or EC* algorithm.
        If registry.settings.jwtauth.private_key is not set,
        registry.settings.jwtauth.master_secret is used.
        registry.settings.jwtauth.algorithm is used as algorithm.

        :param claims_set: set of claims, which will be included in
            the created token.
        """
        token = jwt.encode(
            claims_set, self.private_key, self.algorithm,
        ).decode(encoding="UTF-8")

        return token

    def get_userid(self, claims_set):
        """Extract the userid from a claims set.

        Returns userid or None if there is none.

        :param claims_set: set of claims, which was included
            in the received token.
        """
        if self.userid_claim in claims_set:
            return claims_set[self.userid_claim]
        else:
            return None

    def get_extra_claims(self, claims_set):
        """Get claims holding extra identity info from the claims set.

        Returns a dictionary of extra claims or None if there are none.

        :param claims_set: set of claims, which was included in the received
        token.
        """
        reserved_claims = (
            self.userid_claim,
            "iss",
            "aud",
            "exp",
            "nbf",
            "iat",
            "jti",
            "refresh_until",
            "nonce",
        )
        extra_claims = {}
        for claim in claims_set:
            if claim not in reserved_claims:
                extra_claims[claim] = claims_set[claim]
        if not extra_claims:
            return None

        return extra_claims

    def get_jwt(self, request):
        """Extract the JWT token from the authorisation header of the request.

        Returns the JWT token or None, if the token cannot be extracted.

        :param request: request object.
        :type request: :class:`morepath.Request`
        """
        try:
            authorization = request.authorization
        except ValueError:  # pragma: no cover
            return None
        if authorization is None:
            return None
        authtype, token = authorization
        if authtype.lower() != self.auth_header_prefix.lower():
            return None
        return token

    def verify_refresh(self, request):
        """
        Verify if the request to refresh the token is valid.
        If valid it returns the userid which can be used to create
        an updated identity with ``remember_identity``.
        Otherwise it raises an exception based on InvalidTokenError.

        :param request: current request object
        :type request: :class:`morepath.Request`
        :returns: userid
        :raises: InvalidTokenError, ExpiredSignatureError, DecodeError,
            MissingRequiredClaimError
        """
        if not self.allow_refresh:
            raise InvalidTokenError("Token refresh is disabled")

        token = self.get_jwt(request)
        if token is None:
            raise InvalidTokenError("Token not found")

        try:
            claims_set = self.decode_jwt(
                token, self.verify_expiration_on_refresh
            )

        # reraise the exceptions to change the error messages
        except DecodeError:
            raise DecodeError("Token could not be decoded")
        except ExpiredSignatureError:
            raise ExpiredSignatureError("Token has expired")

        userid = self.get_userid(claims_set)
        if userid is None:
            raise MissingRequiredClaimError(self.userid_claim)

        if self.refresh_nonce_handler is not None:
            if "nonce" not in claims_set:
                raise MissingRequiredClaimError("nonce")
            if (
                self.refresh_nonce_handler(request, userid)
                != claims_set["nonce"]
            ):
                raise InvalidTokenError("Refresh nonce is not valid")

        if self.refresh_delta is not None:
            if "refresh_until" not in claims_set:
                raise MissingRequiredClaimError("refresh_until")
            now = timegm(datetime.utcnow().utctimetuple())
            refresh_until = int(claims_set["refresh_until"])
            if refresh_until < (now - self.leeway):
                raise ExpiredSignatureError("Refresh nonce has expired")

        return userid



from .main import JWTIdentityPolicy


def verify_refresh_request(request):
    """
    Wrapper around JWTIdentityPolicy.verify_refresh which verify
    if the request to refresh the token is valid.
    If valid it returns the userid which can be used to create to
    create an updated identity with ``remember_identity``.
    Otherwise it raises an exception based on InvalidTokenError.

    :param request: request object
    :type request: :class:`morepath.Request`
    :returns: userid
    :raises: InvalidTokenError, ExpiredSignatureError, DecodeError,
        MissingRequiredClaimError
    """
    jwtauth_settings = request.app.settings.jwtauth.__dict__.copy()
    identity_policy = JWTIdentityPolicy(**jwtauth_settings)

    return identity_policy.verify_refresh(request)
